/*global Util, Query, angular, Configuration*/
var TcoUtil = function(ScreenService, General, ValidationEngine, eventAggregator, ZHPromise, ZhPreferences, ZhFilterPreferences, maskEngine, templateManager, ZhKeyboardService, metaDataFactory, MessageService) {

    this.validations = function() {

        var codeValidation = function(value, validationAttrs) {
            var retorno;

            retorno = {
                "valid": (!isNaN(value)),
                "message": ScreenService.i18n(validationAttrs.message) || ScreenService.i18n("Código inválido")
            };

            return retorno;
        };

        ValidationEngine.registerValidation('isValidCodeNumber', codeValidation);
    };

    eventAggregator.subscribe('authSuccess', function(session) {
        this.validations();
    }.bind(this));

    this.delimitCaracter = function(args, limit) {
        if (!Util.isEmptyOrBlank(args.owner.field.value())) {
            var ano = args.owner.field.value().toString();
            if (ano.length > limit) {
                args.owner.field.setValue(parseInt(ano.substr(0, limit), 10));
            }
        }
    };

    this.startFilter = function(widget, currentRow) {
        return ZhPreferences.getDefaultWidgetPreferences(widget.id).then(function(result) {
            if (result && result.DEFAULT_FILTER) {
                ZhPreferences.getPreferences('filter', Query.build().where('WIDGET_ID').equals(widget.id)).then(function(filters) {
                    var defaultFilter = filters.filter(function(filter) {
                        return filter.ID == result.DEFAULT_FILTER;
                    })[0];
                    var filter = {
                        ID: defaultFilter.LABEL_TO_PREFERENCE || 'EDITED',
                        FILTER_ROW: JSON.parse(defaultFilter.FILTER_ROW)
                    };

                    currentRow = filter.FILTER_ROW;

                    ZhFilterPreferences.applyFilter(widget, filter).then(function() {
                        widget.appliedFilter = widget.appliedFilter || filter;
                        widget.appliedFilter.ID = result.DEFAULT_FILTER;
                        if (defaultFilter.LABEL_TO_PREFERENCE) {
                            widget._filterInfo[0].value = defaultFilter.LABEL_TO_PREFERENCE;
                        }
                    });
                });
            } else {
                if (Util.isUndefinedOrNull(currentRow) || Util.isEmptyObject(currentRow)) {
                    currentRow = buildFilterCurrentRow(widget);
                }
            }
            ZhFilterPreferences.openFilterWidget(widget, currentRow);
        });
    };

    var buildFilterCurrentRow = function(widget) {
        // let filterFields = getFilterFields(widget);
        return getFilterRowWithDefaultValues(widget.fields);
    };

    var getFilterRowWithDefaultValues = function(filterFields) {
        let filterRow = {};
        filterFields.forEach(field => {
            let defaultValue = Util.getNestedProperty(field, ["filterProperties", "defaultValue"]) || field.defaultValue;
            if (!Util.isUndefinedOrNull(defaultValue)) {
                filterRow[field.name] = Util.evaluate(defaultValue);
            }
        });
        return filterRow;
    };

    this.justInteger = function(selector, retryCount) {
        if (!retryCount || retryCount > 5)
            retryCount = 5;
        if (!selector) {
            throw {
                "message": "Selector is not defined!"
            };
        }
        setTimeout(function() {
            var elem = $(selector);
            if (elem[0]) {
                var length = selector.split(',').length;
                for (var i = 0; i < length; i++) {
                    if (elem.hasOwnProperty(i)) {
                        elem[i].onkeypress = avoidNonInteger;
                        elem[i].onpaste = preventEvent;
                    }
                }
            } else {
                if (!retryCount) {
                    console.error("Selector not found!");
                } else {
                    retryCount--;
                    this.justInteger(selector, retryCount);
                }
            }
        }.bind(this), 500);
    };

    function avoidNonInteger(e) {
        var integerNumbers = [44, 46, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57];
        var numberCode = e.which;
        if (integerNumbers.indexOf(numberCode) < 0) {
            preventEvent(e);
        }
    }

    function preventEvent(e) {
        e.preventDefault();
    }

    /**
     * Função responsável por chamar a maskCurrency sempre que o 
     * usuário desfocar do campo onde ela é aplicada.
     * Função deve ser usada no evento FieldOnBlur, junto com a 
     * removeMask, essa no FieldOnFocus.
     * 
     * @param {Field} field - field.
     * @param {Char} decimal - Caráctere que será usado para a casa decimal.
     * @param {Char} thousand - Caráctere que será usado para as unidades de milhar.
     * @param {Number} precision - Quantidade de casas decimais.
     */

    this.callMaskCurrency = function(field, decimal, thousands, precision = 2) {
        var value = field.value();
        var params = {
            "decimal": decimal,
            "thousands": thousands,
            "precision": precision
        };
        // field.modelValue();
        // field.widget.redraw();
        // templateManager.updateTemplate()
        field.value(this.maskCurrency(value, params));
    };

    var maskFunction = function(element, viewValue, params, callBack) {
        // do something
    };

    /**
     * Função responsável por aplicar uma máscara do type currency
     * em algum field.
     * 
     * @param {Value} value - value do field.
     * @param {Object} params - Objeto param da máscara com {decimal, thousand, precision}.
     * 
     */

    this.maskCurrency = function(value, params) {
        var thousands = Util.isEmptyOrBlank(params.thousands) ? params.thousand : params.thousands;
        if (!Util.isEmptyOrBlank(value)) {

            if (value[0] === '.') {
                value = '0' + '.' + value.split('.').join('');
            } else if (value[0] === ',') {
                value = '0' + ',' + value.split(',').join('');
            }

            if (params.decimal === ',')
                value = numberFormatBr(value);
            var newValue = "";
            value = value.split('.').join('');
            var dec = value.split(',')[1];
            var mil = [];
            var int = value.split(',')[0];

            if (dec === undefined) {
                dec = "";
                dec = dec.padEnd(params.precision, "0");
                int = value;
            } else if (dec.length < params.precision) {
                dec = dec.padEnd(params.precision, "0");
            } else if (dec.length > params.precision) {
                dec = parseFloat('1.' + dec).toFixed(params.precision).toString().split('.')[1];
            }

            int = parseInt(int).toString();
            var counter = angular.copy(int.length / 3);
            for (var i = 1; i <= counter; i++) {
                if (i != counter)
                    mil.push(thousands + int.substr(-3));
                else
                    mil.push(int.substr(-3));
                int = int.substr(0, (int.length - 3));
            }

            if (int != "") mil.push(int);

            mil = mil.reverse().join('');

            newValue = mil + params.decimal + dec;
            var regex = '^(^([0-9]{1,3}\\' + thousands + ')((\\d{3}\\' + thousands + '{1})|(\\d{3}\\' + params.decimal + '\\d{2,}))*$)|^(\\d{0,3}\\' + params.decimal + '\\d{2,})*$';
            regex = new RegExp(regex, 'g');
            if (regex.test(newValue))
                return newValue;
            else
                return 0;
        }
    };

    maskEngine.registerMask('maskCurrencyTCO', maskFunction, this.maskCurrency);

    /**
     * Função responsável por retirar os pontos da máscara no
     * field para o usuário digitar.
     * 
     * @param {Args} args - Args do field.
     */

    this.removeMask = function(args) {
        var field = args.owner.field;
        var value = field.value();
        if (/^(.*)(\.+).*(,+)(.*)$/gm.test(field.value())) {
            field.value(value.split('.').join(''));
        }
    };

    /**
     * Função que retira os '.' das casas de milhar dos números
     * e troca as ',' dos decimais por '.'
     * Dessa forma o campo fica compativel com o BD do oracle.
     * 
     * @param {String} value - Valor numérico.
     */

    this.parseFloat = function(value) {
        if (/^(.*)(\.+).*(,+)(.*)$/gm.test(value)) {
            var loop = value.toString().split('.');
            for (var i = 0; i < loop.length - 1; i++) {
                value = value.toString().replace('.', '');
            }
        }
        if (/\,/g.test(value))
            value = value.toString().replace(',', '.');
        return parseFloat(value);
    };

    /**
     * Função que redireciona para outra tela fazendo o uso do window history.
     * 
     * @param {String} urlString - URL destino.
     * @param {Object} params - Parametros que podem ser capturados via window.history.
     * @param {String} title - Define um título para o histórico no navegador.
     */

    this.openWindowByHistory = function(urlString, params, title) {
        title = title === null ? "TecFood" : title;
        window.history.pushState(params, title, urlString);
    };

    /**
     * Função responsável por definir os 
     * parâmetros paramsNewCode e __is_new de um 
     * registro antes de salvar.
     * 
     * @param {Widget} widget Widget contendo o registro.
     */
    this.setNewCodeParametersInRow = function(widget) {

        widget.currentRow.__using_unit = widget.paramsNewCode.__using_unit || false;
        widget.currentRow.__new_code_pk = widget.paramsNewCode.__new_code_pk;
        widget.currentRow.__table_pk = widget.paramsNewCode.__table_pk;
        widget.currentRow.__new_code_length = widget.paramsNewCode.__new_code_length || 5;
        widget.currentRow.__new_code_column_size = widget.paramsNewCode.__new_code_column_size || 10;
        widget.currentRow.__cast_number = widget.paramsNewCode.__cast_number || false;

        if (!widget.currentRow.__columns_to_return && widget.paramsNewCode && widget.paramsNewCode.__columns_to_return)
            widget.currentRow.__columns_to_return = widget.paramsNewCode.__columns_to_return;

    };

    /**
     * Preenche a string passada por parâmetro
     * com '0' a esquerda até atingir o tamanho 
     * indicado pelo segundo parâmetro da função.
     * 
     * @param {String} value Valor a ser preenchido
     *  com zerios.
     * @param {Number} length Tamanho total final 
     *  da string.
     * 
     * @return {String} String formatada com '0'.
     */
    this.padLeft = function(value, length) {
        return (value.toString().length < length) ? this.padLeft("0" + value, length) : value;
    };

    /**
     * Preenche a string passada por parâmetro
     * com '0' a esquerda até atingir o tamanho 
     * indicado pelo segundo parâmetro da função.
     * 
     * @param {String} value Valor a ser preenchido
     *  com zerios.
     * @param {Number} length Tamanho total final 
     *  da string.
     * 
     * @return {String} String formatada com '0'.
     */
    this.padRight = function(value, length) {
        return (value.toString().length < length) ? this.padRight(value + "0", length) : value;
    };

    this.buildRegExp = function(value) {
        var filterValue = "^" + value + "$";
        var rgx = /\%/;
        while (rgx.test(filterValue)) {
            filterValue = filterValue.replace(rgx, '(.*?)');
        }
        return new RegExp(filterValue, 'i');
    };

    this.concatenaSemanaEDias = function(args, widget = null) {
        var widgetMain = widget ? widget : args.owner;
        widgetMain.dataSource.data.forEach(function(row, count) {
            if (row.NRSEMANA !== 0)
                widgetMain.dataSource.data[count].LABELDIA = row.NRSEMANA + "° " + ScreenService.i18n(row.LABELDIA);
        });
    };

    this.buildMessagesToShow = function(message) {
        if (typeof message === "string") {

            return MessageService.getMessageByCode(message);

        } else if (typeof message === "object") {
            var mensagens = [];
            var msgParams = [];

            message.forEach(function(msg) {
                msgParams.push(msg.PARAMS);
                mensagens.push(msg.MSGCODE);
            });

            return MessageService.getMessagesByCode(mensagens).then(function(mensagensTraduzidas) {
                for (let i = 0; i < mensagensTraduzidas.length; i++) {
                    mensagensTraduzidas[i] = ScreenService.i18n(mensagensTraduzidas[i], msgParams[i]);
                }
                return mensagensTraduzidas;
            });
        }
    };

    this.buildRegExp = function(value) {
        var filterValue = "^" + value + "$";
        var rgx = /\%/;
        while (rgx.test(filterValue)) {
            filterValue = filterValue.replace(rgx, '(.*?)');
        }
        return new RegExp(filterValue, 'i');
    };

    var numberFormatBr = function(value) {
        if (/^(.*)(\.+).*(\,+)(.*)$/gm.test(value)) {
            let loop = value.toString().split('.');
            for (let i = 0; i < loop.length - 1; i++) {
                value = value.toString().replace('.', '');
            }
        } else if (/^(.*)(\,+).*(\.+)(.*)$/gm.test(value)) {
            let loop = value.toString().split(',');
            for (let i = 0; i < loop.length - 1; i++) {
                value = value.toString().replace(',', '');
            }
        }
        if (/^\d{1,}\.{1}\d*$/g.test(value))
            value = value.toString().replace('.', ',');
        return value.toString();
    };

    /**
     * Função responsável buscar o nivel maximo do Prato.
     * 
     * @param {nameWid} nameWid - Nome da widget que contem o field CDPRATO.
     */

    //CDPICTPRAT
    this.maxLevelPrato = function(nameWid) {
        var widget = templateManager.container.getWidget(nameWid);
        this.getMaxLevelPrato().then(function(response) {
            widget.maxNivelPrato = response.dataset.getMaxLevelPrato.MAXLEVELPRATO;
        });
    };

    this.getMaxLevelPrato = function() {
        return General.callRestEngine('tco#/getMaxLevelPrato', "DataSet", []);
    };

    /**
     * Função responsável buscar o nivel maximo do Produto.
     * 
     * @param {nameWid} nameWid - Nome da widget que contem o field CDARVPROD.
     */

    //CDPICTPROD
    this.maxLevelProduto = function(nameWid) {
        var widget = templateManager.container.getWidget(nameWid);
        this.getMaxLevelProduto().then(function(response) {
            widget.maxNivelProd = response.dataset.getMaxLevelProduto.MAXLEVELPROD;
        });
    };

    this.getMaxLevelProduto = function() {
        return General.callRestEngine('tco#/getMaxLevelProduto', "DataSet", []);
    };

    /**
     * Função responsável buscar a máscara de Produto.
     * 
     * @param {nameWid} nameWid - Nome da widget que contem o field CDARVPROD.
     */

    //CDPICTPROD
    this.maskProduto = function(nameWid) {
        var widget = templateManager.container.getWidget(nameWid);
        this.getMaskProduto().then(function(response) {
            widget.fields.forEach(function(field) {
                if (field.name === "CDARVPROD") {
                    applyMaskProd(field, response);
                }
                if (field.fields) {
                    field.fields.forEach(function(field) {
                        if (field.name === "CDARVPROD") {
                            applyMaskProd(field, response);
                        }
                    });
                }
            });
        });
    };

    this.getMaskProduto = function() {
        return General.callRestEngine('tco#/getMaskProduto', "DataSet", []);
    };

    /**
     * Função responsável buscar a máscara de Prato.
     * 
     * @param {nameWid} nameWid - Nome da widget que contem o field CDPRATO.
     */

    //CDPICTPRAT
    this.maskPrato = function(nameWid) {
        var widget = templateManager.container.getWidget(nameWid);
        this.getMaskPrato().then(function(response) {
            widget.fields.forEach(function(field) {
                if (field.name === "CDPRATO") {
                    applyMaskPrato(field, response);
                }
                if (field.fields) {
                    field.fields.forEach(function(field) {
                        if (field.name === "CDPRATO") {
                            applyMaskPrato(field, response);
                        }
                    });
                }
            });
        });
    };

    this.getMaskPrato = function() {
        return General.callRestEngine('tco#/getMaskPrato', "DataSet", []);
    };

    var applyMaskProd = function(field, response) {
        field.mask = {
            "type": "fix",
            "params": {
                "mask": response.dataset.getMaskProduto.CDPICTPROD.replace(/[a,9]/g, '*')
            }
        };
    };

    var applyMaskPrato = function(field, response) {
        field.mask = {
            "type": "fix",
            "params": {
                "mask": response.dataset.getMaskPrato.CDPICTPRAT.replace(/[a,9]/g, '*')
            }
        };
    };

    /**
     * Função responsável por alterar o estado required 
     * do field passado por parâmetro, corrigindo o 
     * erro de 'required':false do framework.
     * 
     * @params {Field} field - Field a ser alterado.
     * @params {Boolean} state - Estado de required.
     */
    this.changeRequiredField = function(field, state) {
        if (typeof field.validations == 'undefined') {
            field.validations = {};
        }

        if (state) {
            field.validations.required = true;
        } else if (typeof field.validations.required != 'undefined') {
            delete field.validations.required;
        }
    };

    var trataNumerico = function(value, decimalSeparator) {
        if (typeof value == "number") {
            return value.toString().split(".").join(decimalSeparator);
        }
        return (value) ? value.toString() : value + '';
    };

    /*
     * Usar no FieldOnKeydown
     *
     */

    this.lengthCurrency = function(args) {
        var value;
        var fieldValue;
        var field = args.owner.field;
        // var allowedKeys = [8, 9, 37, 38, 39, 40, 46, 44, 108, 188, 194];
        var allowedKeys = ['Tab', 'ArrowLeft', 'ArrowRight', 'ArrowUp', 'ArrowDown', 'End', 'Home', 'Shift', 'Control', 'Delete', 'Backspace', 'Alt'];

        let decimalSeparator = ',';
        let thousandSeparator = '.';

        if (field.mask && field.mask.params && field.mask.params.decimal) {
            decimalSeparator = field.mask.params.decimal;
        }

        if (field.inlineEditable && (field.widget.mode === "list" || field.widget.mode === "inline_edit")) {
            fieldValue = trataNumerico(field.widget.currentRow[field.name + '_inline'], decimalSeparator);
            fieldValue = fieldValue.toString();
            if (event.target.selectionStart !== event.target.selectionEnd) {
                fieldValue = fieldValue.substring(0, event.target.selectionStart) || fieldValue.substring(event.target.selectionEnd);
            }
            value = Util.isEmptyOrBlank(fieldValue) ? "" : fieldValue;
        } else {
            fieldValue = trataNumerico(field.value(), decimalSeparator);
            fieldValue = fieldValue.toString();
            if (event.target.selectionStart !== event.target.selectionEnd) {
                fieldValue = fieldValue.substring(0, event.target.selectionStart) || fieldValue.substring(event.target.selectionEnd);
            }
            value = Util.isEmptyOrBlank(fieldValue) ? "" : fieldValue;
        }

        fieldValue = fieldValue.toString();
        if (event.target.selectionStart !== event.target.selectionEnd) {
            fieldValue = fieldValue.substring(0, event.target.selectionStart) || fieldValue.substring(event.target.selectionEnd);
        }

        let posInsertChar = event.target.selectionStart;

        var cleanNumber = value.split('-').join('').split('.').join('').split(',').join('');
        var realLength = value.split('-').join('').split('.').join('').split(',')[0].length;
        var last = Math.max(...[value.lastIndexOf(','), value.lastIndexOf('.')]);
        var separador;
        var decimalLength = value.split('-').join('').split('.').join('').split(',')[1] ? value.split('-').join('').split('.').join('').split(',')[1].length : 0;
        var precision = "d{0,5}";
        var precisionLength = 0;
        var regex = "^(\\d{1," + field.maxLengthCurrency + "})$";

        if (field.mask && field.mask.params && field.mask.params.precision) {
            precision = "d{1," + field.mask.params.precision + "}";
            regex = "^(\\d{1," + field.maxLengthCurrency + "})(\\,|\\.){1}(\\" + precision + ")$";
            precisionLength = field.mask.params.precision;
        }
        var validation = new RegExp(regex, "gm");

        if (last > -1) {
            separador = value.substring(0, last).length >= field.maxLengthCurrency - precisionLength ? ',' : value[last];
            value = value.substr(0, last).concat(separador + value.substr(last + 1, value.length));
            value = value.split('.').join('');
            last = value.lastIndexOf(',');
            realLength = value.substr(0, last).length;
            decimalLength = value.substr(last + 1, value.length).length;
        }

        let isDecimalSeparator = (args.e.key === decimalSeparator);
        let allowDecimalSeparator = ((precisionLength > 0) && (!value.includes(decimalSeparator)));
        let posDecimalseparator = fieldValue.indexOf(decimalSeparator);
        if (isDecimalSeparator && !allowDecimalSeparator) {
            args.e.preventDefault();
        }

        if (field.mask && field.mask.params && field.mask.params.thousands) {
            thousandSeparator = field.mask.params.thousands;
        }
        let isThousandSeparator = (args.e.key === thousandSeparator);

        if (isThousandSeparator) {
            args.e.preventDefault();
        }

        let allowNegative = true;
        if (field.mask && field.mask.params) {
            allowNegative = field.mask.params.allowNegative;
        }
        allowNegative = (allowNegative && !value.includes('-'));
        isNegativeSimbol = (args.e.key === '-');

        if (isNegativeSimbol && (!allowNegative || (event.target.selectionStart > 0))) {
            args.e.preventDefault();
        }
        if (!allowedKeys.includes(args.e.key) && !isNegativeSimbol && !isDecimalSeparator && !isThousandSeparator) {
            if (realLength >= field.maxLengthCurrency) {
                if ((posDecimalseparator < 0) || (posInsertChar <= posDecimalseparator))
                    args.e.preventDefault();
            }

            if ((posDecimalseparator >= 0) && (posInsertChar > posDecimalseparator) && (precisionLength == decimalLength)) {
                args.e.preventDefault();
            }
        }
    };

    //Crosstab navigation
    var _isKeysBinded = false;
    var kbleft;
    var kbup;
    var kbright;
    var kbdown;
    var kbenter;
    var kbctrlc;
    var kbctrlv;
    var kbctrldel;
    var _gridElement;
    var _selectedCellElement;
    var _selectedRowElement;
    var _copiedCellElement;
    var _copiedRowElement;
    var _rowHeight = 36;
    var crosstabWidget;
    var pasteFunction;
    var deleteFunction;

    this.bindKeyboardEvents = function(funcToPaste, funcToDelete) {
        this.unbindKeyboardEvents();

        kbleft = ZhKeyboardService.bind("left", navigateLeft);
        kbup = ZhKeyboardService.bind("up", navigateUp);
        kbright = ZhKeyboardService.bind("right", navigateRight);
        kbdown = ZhKeyboardService.bind("down", navigateDown);
        kbenter = ZhKeyboardService.bind("enter", openCell);

        if (funcToPaste && typeof funcToPaste === "function") {
            kbctrlc = ZhKeyboardService.bind("mod+c", copyCell);
            kbctrlv = ZhKeyboardService.bind("mod+v", pasteCell);
            pasteFunction = funcToPaste;
        }
        if (funcToDelete && typeof funcToDelete === "function") {
            kbctrldel = ZhKeyboardService.bind("del", deleteCell);
            deleteFunction = funcToDelete;
        }
    };

    this.unbindKeyboardEvents = function() {
        if (kbleft)
            ZhKeyboardService.unbind(kbleft);
        if (kbup)
            ZhKeyboardService.unbind(kbup);
        if (kbright)
            ZhKeyboardService.unbind(kbright);
        if (kbdown)
            ZhKeyboardService.unbind(kbdown);
        if (kbenter)
            ZhKeyboardService.unbind(kbenter);
        if (kbctrlc)
            ZhKeyboardService.unbind(kbctrlc);
        if (kbctrlv)
            ZhKeyboardService.unbind(kbctrlv);
        if (kbctrldel)
            ZhKeyboardService.unbind(kbctrldel);
    };

    function initCrosstabNavigation() {
        if (canMoveCell()) {
            var firstRow = getFirstRowElement(_gridElement);
            var firstCell = getFirstCellElement(firstRow);
            firstCell.classList.add('selected-cell');
            _selectedCellElement = firstCell;
            _selectedRowElement = firstRow;
            _selectedRowElement._childIndex = getElementChildIndex(_selectedRowElement);
        }
    }

    function resetCrosstabNavigation() {
        if (_copiedCellElement && _copiedCellElement.classList) {
            _copiedCellElement.classList.remove('copied-cell');
        }
        _gridElement = false;
        _copiedCellElement = false;
        _copiedRowElement = false;
    }

    function restoreCrosstabNavigation() {
        _gridElement = _gridElement || document.getElementsByClassName('crosstab-body-content')[0];
        if (_gridElement && _selectedRowElement && _selectedCellElement) {
            var rowElementIndex = _selectedRowElement._childIndex;
            var cellElementIndex = getElementChildIndex(_selectedCellElement);
            if (_gridElement.children && _gridElement.children[rowElementIndex]) {
                _selectedRowElement = _gridElement.children[rowElementIndex];
                _selectedRowElement._childIndex = getElementChildIndex(_selectedRowElement);
                if (_selectedRowElement.children && _selectedRowElement.children[cellElementIndex]) {
                    _selectedCellElement = _selectedRowElement.children[cellElementIndex];
                    _selectedCellElement.classList.add('selected-cell');
                    restoreScrollPosition();
                }
            }
        }
    }

    function restoreScrollPosition() {
        var crosstabElement = document.getElementById('grid-' + crosstabWidget.id);
        var XScrollControl = crosstabElement.getElementsByClassName('body-scroll-control')[0];
        if (XScrollControl && _selectedCellElement) {
            var lastPosition = XScrollControl.scrollLeft;
            setTimeout(() => {
                XScrollControl.scrollLeft = lastPosition;
            }, 200);
        }
    }

    function deleteCell() {
        if (_selectedRowElement && _selectedCellElement && canDeleteCell(_selectedCellElement)) {
            ScreenService.confirmMessage('Confirmar exclusão?', 'question', function() {
                var cellIndex = getElementChildIndex(_selectedCellElement);
                var rowCelula = getDataRowByIndex(crosstabWidget.crosstabData, _selectedRowElement._childIndex, cellIndex, crosstabWidget.crosstabColumns);
                deleteFunction(rowCelula);
            });
        }
    }

    function canCopyCell(cell) {
        if (canMoveCell()) // crosstab widget activated
            if (!Util.isEmptyOrBlank(cell.innerText)) // has content on cell 
                return true;

        return false;
    }

    function canDeleteCell(cell) {
        if (canMoveCell()) // crosstab widget activated
            if (getElementChildIndex(cell) > 0)
                if (!Util.isEmptyOrBlank(cell.innerText)) // has content on cell
                    return true;

        return false;
    }

    function canPasteCell(cell) {
        if (canMoveCell()) // crosstab widget activated
            if (getElementChildIndex(cell) > 0)
                return true;

        return true;
    }

    function copyCell() {
        if (_selectedRowElement && _selectedCellElement && canCopyCell(_selectedCellElement)) {
            if (_copiedCellElement) {
                _copiedCellElement.classList.remove('copied-cell');
            }
            _copiedRowElement = _selectedRowElement;
            _copiedCellElement = _selectedCellElement;
            _copiedCellElement.classList.add('copied-cell');
            ScreenService.notificationMessage("Prato copiado.", "success");
        }
    }

    function pasteCell() {
        if (_copiedRowElement && _copiedCellElement && canPasteCell(_selectedCellElement)) {
            var origCellIndex = getElementChildIndex(_copiedCellElement);
            var destCellIndex = getElementChildIndex(_selectedCellElement);
            var origRow = getDataRowByIndex(crosstabWidget.crosstabData, _copiedRowElement._childIndex, origCellIndex, crosstabWidget.crosstabColumns);
            var destRow = getDataRowByIndex(crosstabWidget.crosstabData, _selectedRowElement._childIndex, destCellIndex, crosstabWidget.crosstabColumns);
            pasteFunction(origRow, destRow);
            _copiedCellElement.classList.remove('copied-cell');
        }
    }

    function getDataRowByIndex(data, rowIndex, cellIndex, columns) {
        var row = false;
        if (data[rowIndex]) {
            for (let i = 0; i < Object.keys(data[rowIndex]).length; i++) {
                if (i === cellIndex - 1) {
                    row = data[rowIndex][columns[cellIndex].key].__row;
                }
            }
        }
        return row;
    }

    function canMoveCell() {
        return templateManager.container.getActiveWidget().name == crosstabWidget.name && !ScreenService.isPopupOpen();
    }

    function navigateLeft() {
        _gridElement = _gridElement || document.getElementsByClassName('crosstab-body-content')[0];
        if (!_selectedCellElement) {
            initCrosstabNavigation();
        } else {
            if (canMoveCell()) {
                if (_selectedCellElement.previousElementSibling && !_selectedCellElement.previousElementSibling.classList.contains('fixed')) {
                    _selectedCellElement.classList.remove('selected-cell');
                    _selectedCellElement = _selectedCellElement.previousElementSibling;
                    _selectedCellElement.classList.add('selected-cell');
                    scrollLeft();
                }
            }
        }
    }

    function navigateUp() {
        _gridElement = _gridElement || document.getElementsByClassName('crosstab-body-content')[0];
        if (!_selectedCellElement) {
            initCrosstabNavigation();
        } else {
            if (canMoveCell()) {
                if (_selectedRowElement.previousElementSibling) {
                    var childIndex = getElementChildIndex(_selectedCellElement);
                    _selectedRowElement = _selectedRowElement.previousElementSibling;
                    _selectedCellElement.classList.remove('selected-cell');
                    _selectedRowElement._childIndex = getElementChildIndex(_selectedRowElement);
                    _selectedCellElement = _selectedRowElement.children[childIndex];
                    _selectedCellElement.classList.add('selected-cell');
                    scrollUp();
                }
            }
        }
    }

    function navigateRight() {
        _gridElement = _gridElement || document.getElementsByClassName('crosstab-body-content')[0];
        if (!_selectedCellElement) {
            initCrosstabNavigation();
        } else {
            if (canMoveCell()) {
                if (_selectedCellElement.nextElementSibling) {
                    _selectedCellElement.classList.remove('selected-cell');
                    _selectedCellElement = _selectedCellElement.nextElementSibling;
                    _selectedCellElement.classList.add('selected-cell');
                    scrollRight();
                }
            }
        }
    }

    function navigateDown() {
        _gridElement = _gridElement || document.getElementsByClassName('crosstab-body-content')[0];
        if (!_selectedCellElement) {
            initCrosstabNavigation();
        } else {
            if (canMoveCell()) {
                if (_selectedRowElement.nextElementSibling) {
                    var childIndex = getElementChildIndex(_selectedCellElement);
                    _selectedRowElement = _selectedRowElement.nextElementSibling;
                    _selectedRowElement._childIndex = getElementChildIndex(_selectedRowElement);
                    _selectedCellElement.classList.remove('selected-cell');
                    _selectedCellElement = _selectedRowElement.children[childIndex];
                    _selectedCellElement.classList.add('selected-cell');
                    scrollDown();
                }
            }
        }
    }

    function scrollLeft() {
        var crosstabElement = document.getElementById('grid-' + crosstabWidget.id);
        if (crosstabElement) {
            var XScrollControl = crosstabElement.getElementsByClassName('body-scroll-control')[0];
            if (XScrollControl) {
                if (_selectedCellElement.offsetLeft < XScrollControl.offsetLeft) {
                    XScrollControl.scrollLeft -= XScrollControl.offsetLeft - _selectedCellElement.offsetLeft;
                }
            }
        }
    }

    function scrollRight() {
        var crosstabElement = document.getElementById('grid-' + crosstabWidget.id);
        if (crosstabElement) {
            var XScrollControl = crosstabElement.getElementsByClassName('body-scroll-control')[0];
            if (XScrollControl) {
                if (_selectedCellElement.offsetLeft + _selectedCellElement.offsetWidth > XScrollControl.offsetWidth) {
                    XScrollControl.scrollLeft += _selectedCellElement.offsetWidth;
                }
            }
        }
    }

    function scrollDown() {
        var crosstabElement = document.getElementById('grid-' + crosstabWidget.id);
        if (crosstabElement) {
            var YScrollControl = crosstabElement.getElementsByClassName('body')[0];
            if (YScrollControl) {
                if (getElementChildIndex(_selectedRowElement) > Math.round(YScrollControl.offsetHeight / _rowHeight - 1)) {
                    YScrollControl.scrollTop += _rowHeight;
                }
            }
        }
    }

    function scrollUp() {
        var crosstabElement = document.getElementById('grid-' + crosstabWidget.id);
        if (crosstabElement) {
            var YScrollControl = crosstabElement.getElementsByClassName('body')[0];
            if (YScrollControl) {
                if (getElementChildIndex(_selectedRowElement) < Math.round(YScrollControl.scrollTop / _rowHeight)) {
                    YScrollControl.scrollTop -= _rowHeight;
                }
            }
        }
    }

    function getElementChildIndex(element) {
        if (element) {
            for (var i = 0;
                (element = element.previousElementSibling); i++);
            return i;
        } else {
            return false;
        }
    }

    function openCell() {
        _gridElement = _gridElement || document.getElementsByClassName('crosstab-body-content')[0];
        if (_selectedCellElement && canMoveCell()) {
            _selectedCellElement.click();
        }
    }

    function getFirstRowElement(gridElement) {
        var firstRowElement = false;
        var rowElements = gridElement.children;
        if (rowElements.length) {
            firstRowElement = rowElements[0];
        }
        return firstRowElement;
    }

    function getFirstCellElement(rowElement) {
        var firstCellElement = false;
        var cellElements = rowElement.children;
        if (cellElements.length) {
            var currentElement = cellElements[0];
            while (!firstCellElement && currentElement) {
                if (!currentElement.classList.contains('fixed')) {
                    firstCellElement = currentElement;
                } else {
                    currentElement = currentElement.nextElementSibling;
                }
            }
        }
        return firstCellElement;
    }

    this.updateCustomElements = function(widget) {
        resetCrosstabNavigation();
        if (widget && !widget.defaultReload) {
            widget.defaultReload = widget.reload;
            widget.reload = crosstabCustomReload;
        }
        crosstabWidget = widget;
    };

    function crosstabCustomReload(additionalFilters, shouldRefreshCurrentRow, keepCopiedCell) {
        ScreenService.changeLoadingMessage(ScreenService.i18n("Carregando cardápio..."));
        ScreenService.showLoader();
        return crosstabWidget.defaultReload(additionalFilters, shouldRefreshCurrentRow).then(function() {
            if (!keepCopiedCell) {
                _copiedCellElement = false;
                _copiedRowElement = false;
            }
            setTimeout(function() {
                restoreCrosstabNavigation();
                crosstabWidget.activate();
                ScreenService.hideLoader();
                ScreenService.restoreDefaultLoadingMessage();
            }, 200);
        });
    }

    this.retornaRowSelecionada = function() {
        var cellIndex = getElementChildIndex(_selectedCellElement);
        if (cellIndex) {
            var rowCelula = getDataRowByIndex(crosstabWidget.crosstabData, _selectedRowElement._childIndex, cellIndex, crosstabWidget.crosstabColumns);
            return rowCelula;
        } else {
            return false;
        }
    };

    this.isClickOnFocusedCell = function() {
        var targetElement = window.event.target.closest('.td.zh-standard-column');
        return targetElement && targetElement == _selectedCellElement;
    };

    this.setCurrentNavigationElement = function() {
        if (_selectedCellElement) {
            _selectedCellElement.classList.remove('selected-cell');
        }
        var target = window.event.target;
        var cell = target.closest('.td.zh-standard-column');
        if (cell) {
            var row = cell.closest('.tr');
            if (row) {
                _selectedRowElement = row;
                _selectedRowElement._childIndex = getElementChildIndex(_selectedRowElement);
                _selectedCellElement = cell;
                _selectedCellElement.classList.add('selected-cell');
            }
        }
    };

    var reportWidget = {
        "id": 0,
        "name": "reportWidget",
        "label": "Tipo de Relatório",
        "template": "widget/form.html",
        "container": {},
        "isVisible": true,
        "popupNoBlock": false,
        "fields": [],
        "actions": [],
        "events": []
    };
    var reportFields = [
        {
            "id": 99999999,
            "name": "BIRTTYPE",
            "label": "",
            "class": 12,
            "readOnly": false,
            "isVisible": true,
            "isVisibleForm": true,
            "isVisibleGrid": false,
            "template": "field/radio-group.html",
            "events": [],
            "valueField": "value",
            "descriptionField": "name",
            "defaultOption": 0,
            "dataSource": {
                "data": [
                    {
                        "name": "pdf",
                        "value": "pdf"
                    },
                    {
                        "name": "xlsx",
                        "value": "xlsx"
                    }
                ]
            }
        }
    ];
    var reportActions = [
        {
            "name": "cancelar",
            "label": "Cancelar",
            "idVisible": true,
            "hideIcon": true,
            "showAsAction": "back",
            "events": [
                {
                    "name": "ActionEvent",
                    "code": "ScreenService.closePopup();"
                }
            ]
        },
        {
            "name": "confirmar",
            "label": "Confirmar",
            "idVisible": true,
            "hideIcon": true,
            "showAsAction": "never",
            "events": [
                {
                    "name": "ActionEvent",
                    "code": "TcoUtil.generateReport(args.owner.widget, args);"
                }
            ]
        }
    ];

    var reportCallback;
    var reportParameters;
    var reportWidgetParentName;

    this.openReportPopup = function(widget, callback, parameters) {

        reportCallback = callback;
        reportParameters = parameters || [];
        reportWidgetParentName = widget.name;
        reportWidget.container = widget.container;
        reportWidget.fields = reportFields;
        reportWidget.actions = reportActions;

        var reportWidgetInitialized = metaDataFactory.widgetFactory(reportWidget);

        ScreenService.openPopup(reportWidgetInitialized);
    };

    this.generateReport = function(widget, args) {

        var reportWidgetParentCurrentRow = widget.container.getWidget(reportWidgetParentName).currentRow;

        reportWidgetParentCurrentRow.BIRTTYPE = widget.currentRow.BIRTTYPE;

        args.row = reportWidgetParentCurrentRow;

        if (typeof reportCallback == 'function') {

            reportCallback.apply(null, reportParameters);
        }

        ScreenService.closePopup();

    };

    /*
     * Formata o texto usando máscara separada por pontos e retorna texto formatado
     *
     */

    this.formatString = function(text, mask, removeEmptyPlaceholders = true) {
        if (!text || !mask) return text;
        let newText = text;
        let regex = '';
        let toReplace = '';
        mask.match(/[^\.]+/g).forEach((math, index) => {
            regex += index === 0 ? `(.{${math.length}})` : `(.{0,${math.length}})`;
            toReplace += index !== 0 ? '.' : toReplace;
            toReplace += '$' + (index + 1);
        });
        regex = new RegExp(regex);
        newText = newText.replace(regex, toReplace);
        if (removeEmptyPlaceholders)
            newText = newText.replace(/[.]+$/, '');
        return newText;
    };

    this.clearFilterReport = function(args) {
        args.e.metadata.filter = [];
    };

    function setDefaultValuesConfig(stripeWidget) {
        if (!stripeWidget || stripeWidget == 'DEFAULT') {
            stripeWidget = templateManager.container.widgets.filter(widget =>
                widget.template && widget.template.toLowerCase().indexOf('stripe') > -1
            )[0];
        } else {
            stripeWidget = typeof stripeWidget === 'string' ? templateManager.container.getWidget(stripeWidget) : stripeWidget;
        }
        stripeWidget.floatingControlConfig = stripeWidget.floatingControlConfig || {};
        stripeWidget.floatingControlConfig.dataset = stripeWidget.floatingControlConfig.dataset || [];
        stripeWidget.floatingControlConfig.rangeFields = stripeWidget.floatingControlConfig.rangeFields || [];
        stripeWidget.floatingControlConfig.ignoreFieldsStripe = stripeWidget.floatingControlConfig.hasOwnProperty('ignoreFieldsStripe') ? stripeWidget.floatingControlConfig.ignoreFieldsStripe : false;
        return stripeWidget;
    }

    this.setFilterReport = function(args, stripeWidget) {
        stripeWidget = setDefaultValuesConfig(stripeWidget);
        args.e.metadata.filter = stripeWidget.floatingControlConfig.dataset;
    };

    function treatLabel(label) {
        return ScreenService.i18n(label).split(':').join('');
    }
    /*
        Inserir essa função apos aplicar o filtro, seja no FilterAfterApply ou em um OnChange caso seja uma 
        field que filtra a tela ao inves de um popup de filtro
        
        @params
        stripeWidget => nome da widget de stripe ou o objecto em si
        row (opcional) => row contendo os dados que vão para o filtro do relatorio
        fieldsToUse (opcional) => array com nome de fields a serem exibidas no filtro do relatorio
        ['NMFILIAL']
        default:
            ele exibe todas as fields visiveis do stripe no filtro do relatorio
        Issue:
            stripeWidget por default procurar por uma widget com template widget/stripe.html
    */
    this.newFilterReport = function(stripeWidget, row, fieldsToUse) {
        stripeWidget = setDefaultValuesConfig(stripeWidget);
        let config = stripeWidget.floatingControlConfig;
        let dataSet = [];
        row = row || stripeWidget.currentRow;
        let arrayKeysRangeFields = Object.keys(config.rangeFields);
        let allFieldsRange = arrayKeysRangeFields.join('_');
        if (Array.isArray(arrayKeysRangeFields) && arrayKeysRangeFields.length) {
            arrayKeysRangeFields.forEach(rangeField => {
                let dataSets = {
                    $I: {
                        propertyToGet: 'INDEX',
                        value: config.rangeFields[rangeField].itemsToReplace,
                    },
                    $V: {
                        propertyToGet: 'NAMEFIELD',
                        value: row
                    }
                };
                let fieldsRange = rangeField.split('_');
                let treatStr = config.rangeFields[rangeField].format.replace(/\$[A-Z]+\d+/g, function(currentMatch) {
                    let cKey = currentMatch.match(/(\$[A-Za-z]+)/g)[0];
                    let arrayValues = dataSets[cKey];
                    let index = currentMatch.replace(cKey, '');
                    let currentItem = arrayValues.propertyToGet == 'INDEX' ? arrayValues.value[index] : arrayValues.value[fieldsRange[index]];
                    return ScreenService.i18n(currentItem);
                });
                dataSet.push({
                    label: treatLabel(config.rangeFields[rangeField].label),
                    value: treatStr
                });
            });
        }
        if (Array.isArray(fieldsToUse) && fieldsToUse.length) {
            fieldsToUse.forEach(fieldName => {
                let rangeContainsCurrentField = new RegExp(`(?<!([A-Z]+))(${fieldName})(?!([A-Z]+))`, 'gm').test(allFieldsRange);
                if (!rangeContainsCurrentField) {
                    let field = stripeWidget.getField(fieldName);
                    dataSet.push({
                        label: treatLabel(field.label),
                        value: row[field.name]
                    });
                }
            });
        }
        if (!config.ignoreFieldsStripe) {
            stripeWidget.fields.forEach(field => {
                if (field.isVisible) {
                    dataSet.push({
                        label: treatLabel(field.label),
                        value: row[field.name]
                    });
                }
            });
        }
        config.dataset = dataSet;
    };

};

Configuration(function(ContextRegister, eventEngine) {
    ContextRegister.register('TcoUtil', TcoUtil);
});
